package main

import (
	pb "github.com/paweljw/go-wasm-worker-example/proto"
	"google.golang.org/protobuf/proto"
	"google.golang.org/protobuf/reflect/protoreflect"
	"google.golang.org/protobuf/types/known/timestamppb"
	"log"
	"syscall/js"
	"time"
)

var focused bool
var fibonacciSecondToLast int64
var fibonacciLast int64

// Necessary to pack any array we send to JS into []interface{} first
func marshalMessage(t string, m protoreflect.ProtoMessage) (map[string]interface{}, error) {
	out, err := proto.Marshal(m)
	if err != nil {
		return map[string]interface{}{}, err
	}

	passable := make([]interface{}, len(out))
	for i, v := range out {
		passable[i] = v
	}

	obj := map[string]interface{}{
		"contains": t,
		"payload":  passable,
	}

	return obj, nil
}

func unmarshalMessage(ary js.Value, message proto.Message) error {
	length := ary.Get("length").Int()
	converted := make([]byte, length)
	for i := 0; i < length; i++ {
		converted[i] = byte(ary.Call("at", i).Int())
	}
	return proto.Unmarshal(converted, message)
}

func focusChangeReactor(v js.Value) {
	msg := &pb.FocusChange{}
	err := unmarshalMessage(v, msg)

	if err != nil {
		log.Printf("error unmarshaling FocusChange message: %s", err)
		return
	}

	focused = msg.State == pb.FocusChange_FOCUSED
}

func fibonacciReactor(v js.Value) {
	msg := &pb.Fibonacci{}
	err := unmarshalMessage(v, msg)

	if err != nil {
		log.Printf("error unmarshaling Fibonacci message: %s", err)
		return
	}

	fibonacciLast = msg.Last
	fibonacciSecondToLast = msg.SecondToLast
}

func messageListener(_ js.Value, args []js.Value) interface{} {
	messageType := args[0].Get("contains").String()

	switch messageType {
	case "fibonacci":
		fibonacciReactor(args[0].Get("payload"))
	case "focus_change":
		focusChangeReactor(args[0].Get("payload"))
	}

	return nil
}

func main() {
	focused = false
	js.Global().Set("myListener", js.FuncOf(messageListener))

	fibonacciSecondToLast = 1
	fibonacciLast = 1

	// Say WASM is ready
	msg := &pb.Command{
		Version: 1,
		Type:    pb.Command_READY,
		SentAt:  timestamppb.New(time.Now()),
		Message: "",
	}
	packed, err := marshalMessage("command", msg)
	if err != nil {
		log.Printf("Failed to pack message: %s", err)
	} else {
		js.Global().Call("postMessage", packed)
		time.Sleep(time.Second)
	}

	for {
		for {
			if focused {
				break
			} else {
				time.Sleep(time.Second)
			}
		}

		fibonacciNext := fibonacciLast + fibonacciSecondToLast
		fibonacciSecondToLast = fibonacciLast
		fibonacciLast = fibonacciNext

		msg := &pb.Fibonacci{
			Version:      1,
			SecondToLast: fibonacciSecondToLast,
			Last:         fibonacciLast,
		}

		packed, err := marshalMessage("fibonacci", msg)
		if err != nil {
			log.Printf("Failed to pack message: %s", err)
		} else {
			js.Global().Call("postMessage", packed)
			time.Sleep(time.Second)
		}
	}
}
